require_relative 'player'
require_relative 'notifier'
require_relative 'game_snapshot'
require_relative 'fuse'

module Gameworks
  class Game
    TURN_TIME_LIMIT = 30
    DEFAULT_DELAY = 0.3

    STATE_INITIATING = 'initiating'
    STATE_IN_PLAY = 'in play'
    STATE_COMPLETED = 'completed'

    attr_reader :players, :state, :id, :delay_time, :moves, :notifier

    def initialize(payload={})
      @notifier = Notifier.new
      @players = []
      @state = STATE_INITIATING
      @delay_time = payload['delay_time'] || DEFAULT_DELAY
      @turns = {}
      @observers = {}
      @fuses = {}
      @id = generate_uuid.split('-').first
      @moves = {}
    end

    def register_observer(cb)
      token = generate_uuid
      @observers[token] = [self.snapshot, cb]
      token
    end

    def update_observers(end_game = false)
      @observers.each do |token, (_, cb)|
        snapshot = snapshot_for_observer(token)
        cb.call(delta(snapshot).to_json)
        cb.succeed if end_game
      end
      notifier.push(self, to_json)
    end

    def snapshot_for_observer(token)
      snapshot = nil
      if @observers.has_key?(token)
        snapshot, cb = @observers[token]
        @observers[token][0] = self.snapshot
      end
      snapshot
    end

    def player_changed?(player, snapshot, for_player=nil)
      player.score != snapshot.player_scores[player]
    end

    def delta(snapshot, for_player=nil)
      { :id => id,
        :players => @players.select{ |p| player_changed?(p, snapshot, for_player) }.map{ |p| p.as_json(for_player) },
        :state   => @state }
    end

    def full?
      raise NotImplementedError
    end

    def player_class
      Gameworks::Player
    end

    def start_if_ready
      if full?
        init_game
        signal_turns
        update_observers
      end
    end

    def add_player(payload={})
      return false if full?

      player = player_class.new(payload)
      return false unless player.valid?

      @players << player
      update_observers
      player
    end

    def init_game
      @players.shuffle!
      @state = STATE_IN_PLAY
    end

    def generate_uuid
      SecureRandom.uuid
    end

    def active_players
      raise NotImplementedError
    end

    def ignore_disqualified_players
      @players.rotate! while @players.first.disqualified?
    end

    def signal_turns
      moves.clear
      active_players.each do |player|
        token = generate_uuid
        @turns[token] = player
        # This may no longer be necessary, since we abort the players' fuses
        # when they submit turns in #add_move.
        @fuses[player].abort if @fuses[player]
        @fuses[player] = Gameworks::Fuse.new(TURN_TIME_LIMIT) { disqualify(player) }
        player.signal_turn(token)
      end
    end

    def player_for_token(token)
      @turns.delete(token)
    end

    def add_move(payload, player)
      @fuses[player].abort

      move, error = build_move(payload, player)
      return false, error unless move

      legal, error = move_legal?(move, player)
      return false, error unless legal

      moves[player] = move
      process_move(move, player)

      ignore_disqualified_players
      update_observers
      true
    end

    def build_move(payload, player)
      raise NotImplementedError
    end

    def move_legal?(move, player)
      raise NotImplementedError
    end

    def process_move(move, player)
      raise NotImplementedError
    end

    def waiting_on_moves?
      (active_players - moves.keys).any?
    end

    def disqualify(player)
      player.disqualify!
      update_observers
      if @players.select { |p| !p.disqualified? }.size < 2
        end_game
      elsif !waiting_on_moves?
        signal_turns
      end
      # else do nothing; waiting for other players
    end

    def end_game
      @state = STATE_COMPLETED
      @fuses.values.each(&:abort)
      update_observers(true)
      @players.each{ |p| p.signal_turn(nil) }
    end

    def snapshot_class
      Gameworks::Game::Snapshot
    end

    def snapshot
      snapshot_class.new(self)
    end

    def as_json(for_player=nil)
      { :id      => id,
        :players => @players.map{ |player| player.as_json(for_player) },
        :state   => @state }
    end

    def to_json(for_player=nil)
      as_json(for_player).to_json
    end
  end
end
